'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _cloneDeep = require('lodash/cloneDeep');

var _cloneDeep2 = _interopRequireDefault(_cloneDeep);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Transform operations `a` by operations `b` with `priority`.
 *
 * @param {Array} a
 * @param {Array} b
 * @param {String} priority
 */

function transform(a, b, priority) {
  // Deeply clone A so that the transformers can edit in place without worry.
  a = (0, _cloneDeep2.default)(a);

  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = b[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var bo = _step.value;

      // Create a temporary array to store the new A operations, because sometimes
      // the transformers need to return `null`, to remove an operation.
      var tmp = [];

      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        for (var _iterator2 = a[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          var ao = _step2.value;

          var ret = t(ao, bo, priority);
          if (ret === null) continue;
          tmp.push(ao);
        }

        // Update A for the next iteration.
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }

      a = tmp;
    }

    // Return the transformed A operations.
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  return a;
}

/**
 * Find the place where operation A's path `a` is intersected by operation B's
 * path `b`, if it does at all.
 *
 * @param {Array} a
 * @param {Array} b
 * @return {String}
 */

function intersect(a, b) {
  if (!a || !b) return null;

  for (var i = 0; i < b.length; i++) {
    var bv = b[i];
    var av = a[i];
    var isLast = i + 1 === b.length;
    // If ever B > A, they never intersect.
    if (bv > av) return 'none';
    // If we're not at the end of B, and B < A, they don't intersect.
    if (!isLast && bv < av) return 'none';
    // If we're at the end of B, and B <= A, this is the intersection.
    if (isLast && bv == av) return 'equal';
    if (isLast && bv < av) return 'before';
  }

  return 'none';
}

/**
 * Transform operation `a` in place based on operation `b` and priority `p`.
 *
 * @param {Object} a
 * @param {Object} b
 * @param {String} p
 */

function t(a, b, p) {
  var i = b.path.length - 1;
  var ni = b.newPath ? b.newPath.length - 1 : null;

  var a_b = intersect(a.path, b.path);
  var an_b = intersect(a.newPath, b.path);
  var a_bn = intersect(a.path, b.newPath);
  var an_bn = intersect(a.newPath, b.newPath);

  /**
   * Insert node.
   */

  if (b.type == 'insert_node') {
    // If B's path is before A's, increment.
    if (a_b == 'before') {
      a.path[i]++;
    }

    // If the paths are equal, and B has priority, increment.
    if (a_b == 'equal' && p == 'right') {
      a.path[i]++;
    }

    // If A has a new path, and B's path is before A's, increment.
    if (an_b != null && an_b == 'before') {
      a.newPath[i]++;
    }

    // If A has a new path, and they're equal, and B has priority, incremenet.
    if (an_b != null && an_b == 'equal' && p == 'right') {
      a.newPath[i]++;
    }
  }

  /**
   * Remove node.
   */

  if (b.type == 'remove_node') {
    // If A's and B's paths are equal, and B has priority, ignore operation A.
    if (a_b == 'equal' && p == 'right') {
      return null;
    }

    // If B's path is before A's path, decrement A's.
    if (a_b == 'before') {
      a.path[i]--;
    }

    // If A has a new path, and B's path is before A's new path, decrement A's
    // new path.
    if (an_b != null && an_b == 'before') {
      a.newPath[i]--;
    }

    // If A has a new path, and A's new path and B's path are equal, and B has
    // priority, decrement A's new path.
    if (an_b != null && an_b == 'equal' && p == 'right') {
      a.newPath[i]--;
    }
  }

  /**
   * Move node.
   */

  if (b.type == 'move_node') {
    // If B's path is before A's, and B's new path doesn't intersect, decrement.
    if (a_b == 'before' && a_bn == 'none') {
      a.path[i]--;
    }

    // If B's path is equal to A's, and B has priority, and B's new path doesn't
    // intersect, decrement.
    if (a_b == 'equal' && p == 'right' && a_bn == 'none') {
      a.path[i]--;
    }

    // If A has a new path, and B's path is before A's new path, and B's new
    // path doesn't intersect, decrement.
    if (an_b != null && an_b == 'before' && an_bn == 'none') {
      a.newPath[i]--;
    }

    // If A has a new path, and B's path is equal to A's, and B has priority,
    // and B's new path doesn't intersect, decrement.
    if (an_b != null && an_b == 'equal' && p == 'right' && an_bn == 'none') {
      a.newPath[i]--;
    }

    // If B's new path is before A's, and B's old path doesn't intersect,
    // increment.
    if (a_b == 'none' && a_bn == 'before') {
      a.path[ni]++;
    }

    // If B's new path is equal to A's, and B has priority, and B's old path
    // doesn't intersect, increment.
    if (a_b == 'none' && a_bn == 'equal' && p == 'right') {
      a.path[ni]++;
    }

    // If A has a new path and B's new path is before A's, and B's old path
    // doesn't intersect, increment.
    if (an_b != null && an_b == 'none' && an_bn == 'before') {
      a.newPath[ni]++;
    }

    // If A has a new path and B's new path is equal to A's, and B has priority,
    // and B's old path doesn't intersect, increment.
    if (an_b != null && an_b == 'none' && an_bn == 'equal' && p == 'right') {
      a.newPath[ni]++;
    }
  }
}

/**
 * Export.
 *
 * @type {Function}
 */

exports.default = transform;